package mytype.mycom.mygroup;

import com.tencent.wework.SpecUtil;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.concurrent.TimeUnit;

/** 
 * @usage 提供数据的本地存储功能和过期数据的清理功能
 *        每个记录使用一个文件，文件名为notifyId，文件内容为响应数据，单行
 *        过期文件的定期清理使用独立线程完成，文件默认20分钟过期，每10分钟扫描一次
 */
public class DataBaseUtils {
    /**
     * @usage 存储响应数据的路径，类的静态代码会自动创建
     */
    private static final String DATA_DIR = "/mnt/data/callback_notify/";
    
    /**
     * @usage 清理过期文件的扫描间隔，单位为分钟 
     */
    private static final int SCAN_INTERVAL = 10;  

    /**
     * @usage 文件过期时间，单位为分钟
     */
    private static final int EXPIRE_TIME = 20;  

    /**
     * @usage 1. 创建存放数据的目录
     *        2. 启动清理过期文件的线程
     */
    static {
        Path dirPath = Paths.get(DATA_DIR);
        if (!Files.exists(dirPath)) {
            try {
                Files.createDirectories(dirPath);
            } catch (IOException e) {
                SpecUtil.WWSpecLogError("create directory " + DATA_DIR + " failed");
                e.printStackTrace();
                System.exit(1);
            }
            SpecUtil.WWSpecLogInfo("create directory " + DATA_DIR + " success");
        }

        Thread cleanerThread = new Thread(new Runnable() {
            @Override
            public void run() {
                try {
                    while (true) {
                        File folder = new File(DATA_DIR);
                        File[] files = folder.listFiles();
                        if (files != null) {
                            for (File file : files) {
                                if (file.isFile() && isFileOld(file)) {
                                    SpecUtil.WWSpecLogInfo("Deleted expired file: " + file.getName());
                                    file.delete();
                                }
                            }
                        }
                        TimeUnit.MINUTES.sleep(SCAN_INTERVAL); // 每10分钟执行一次
                    }
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                    SpecUtil.WWSpecLogError("File cleaning thread was interrupted.");
                }
            }

            private boolean isFileOld(File file) {
                try {
                    BasicFileAttributes attrs = Files.readAttributes(file.toPath(), BasicFileAttributes.class);
                    long fileCreationTime = attrs.creationTime().toMillis();
                    return (System.currentTimeMillis() - fileCreationTime) > EXPIRE_TIME * 60 * 1000;
                } catch (Exception e) {
                    SpecUtil.WWSpecLogError("Error reading file attributes: " + e.getStackTrace());
                    return false;
                }
            }
        });

        cleanerThread.setDaemon(true);
        cleanerThread.start();
    }

    public static enum ErrorCode {
        // 成功
        SUCCESS,
        // IO异常
        IO_FAIL,
        // 文件夹不存在
        DIRECTORY_NOT_EXIST,
        // notify id(即文件)不存在
        NOTIFY_ID_NOT_EXIST,
        // 文件为空
        EMPTY_FILE
    }

    /**
     * @usage 存储响应数据
     * @param notifyId 通知ID
     * @param data 响应数据
     * @return ErrorCode 错误码
     * @warning 写入时需要指定为UTF-8编码
     */
    public static ErrorCode setNotifyData(String notifyId, String data) {
        
        try {
            String fileName = DATA_DIR + notifyId;
            File file = new File(fileName);
            if (!file.exists()) {
                file.createNewFile();
                SpecUtil.WWSpecLogInfo("create file '" + fileName + "' success");
            }

            BufferedWriter writer = new BufferedWriter(
                new OutputStreamWriter(
                    new FileOutputStream(fileName), StandardCharsets.UTF_8));
            writer.write(data);
            writer.close();

        } catch (IOException e) {
            e.printStackTrace();
            return ErrorCode.IO_FAIL;
        }

        return ErrorCode.SUCCESS;
    }

    /**
     * @usage 读取指定notify_id的数据
     * @param notifyId 通知ID
     * @param dataBuffer 响应数据缓冲区，用于接收数据
     * @return ErrorCode 错误码
     * @warning 读取时需要指定为UTF-8编码
     */
    public static ErrorCode getByNotifyId(String notifyId, StringBuffer dataBuffer) {
        String fileName = DATA_DIR + notifyId;
        File file = new File(fileName);
        if (!file.exists()) {
            SpecUtil.WWSpecLogError("notify_id not exist");
            return ErrorCode.NOTIFY_ID_NOT_EXIST;
        }

        try {
            BufferedReader reader = new BufferedReader(
                new InputStreamReader(
                    new FileInputStream(fileName), StandardCharsets.UTF_8));
            String line = reader.readLine();
            if (line == null || line.isEmpty()) {
                SpecUtil.WWSpecLogError("file: '" + fileName + "' is empty");
                reader.close();
                reader.close();
                return ErrorCode.EMPTY_FILE;
            }
            dataBuffer.append(line);
            reader.close();
            
        } catch (IOException e) {
            e.printStackTrace();
            return ErrorCode.IO_FAIL;
        }

        return ErrorCode.SUCCESS;
    }

    /**
     * @usage 私有构造函数，防止外部实例化
     */ 
    private DataBaseUtils() {}
}
