package mytype.mycom.mygroup;


import com.tencent.wework.SpecUtil;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.codec.http.*;
import io.netty.channel.socket.SocketChannel;
import java.util.UUID;


/*
 * @usage 网络服务实现类
 */
public class NetworkService {

    /*
     * @description 网络服务监听的IP
     */
    public static final String HOST = "0.0.0.0";

    /*
     * @description 网络服务监听的端口
     */
    public static final int PORT = 8080;

    /*
     * @description DEMO使用的SDK的版本号
     */
    private static final String SDK_VERSION = SpecUtil.GetSDKVersion();

    /*
     * @usage 初始化并启动netty服务
     */
    public static void startServer() throws Exception {

        EventLoopGroup bossGroup = new NioEventLoopGroup(SvrConfig.boss_group_size);
        EventLoopGroup workerGroup = new NioEventLoopGroup(SvrConfig.io_gorup_size);

        try {
            // 初始化服务器
            ServerBootstrap sbs = new ServerBootstrap();
            sbs.group(bossGroup, workerGroup)
                    .channel(NioServerSocketChannel.class)
                    // 设置bossGroup接收连接的队列长度
                    .option(ChannelOption.SO_BACKLOG, 4096)
                    // 地址复用，允许绑定处于TIME_WAIT状态下的socket
                    .option(ChannelOption.SO_REUSEADDR, true)
                    // 禁用Nagle算法，允许立即发送数据包，从而降低延迟
                    .childOption(ChannelOption.TCP_NODELAY, true)
                    // 设置请求处理器的流水线
                    .childHandler(new ChannelInitializer<SocketChannel>() {
                        @Override
                        protected void initChannel(SocketChannel socketChannel) {
                            ChannelPipeline cp = socketChannel.pipeline();
                            // HTTP 解码器：将字节流解码为 HTTP 对象
                            cp.addLast(new HttpServerCodec());
                            // 聚合器：将多个 HTTP 消息部分聚合成完整的 HTTP 消息
                            cp.addLast(new HttpObjectAggregator(64 * 1024 * 1024));
                            // 业务逻辑处理器
                            cp.addLast(new SpecHandler());
                        }
                    });

            // 启动服务器
            Channel ch = sbs.bind(HOST, PORT).sync().channel();
            SpecUtil.WWSpecLogInfo(
                    "server started",
                    "host: " + NetworkService.HOST,
                    "port: " + NetworkService.PORT,
                    "SDK version: " + SDK_VERSION,
                    "boss group size: " + SvrConfig.boss_group_size,
                    "io worker group size: " + SvrConfig.io_gorup_size,
                    "process in io thread:" + SvrConfig.process_in_iothread
            );

            // 启动性能监控线程
            ResourceMonitor monitor = new ResourceMonitor(SvrConfig.monitor_interval);
            monitor.start();

            ch.closeFuture().sync();

        } finally {
            bossGroup.shutdownGracefully();
            workerGroup.shutdownGracefully();
            SpecUtil.WWSpecLogInfo("server shutdown");
        }
    }

    /*
     * @usage 入参解析和发送响应，不涉及具体的业务处理逻辑
     */
    private static class SpecHandler extends SimpleChannelInboundHandler<FullHttpRequest> {

        @Override
        protected void channelRead0(ChannelHandlerContext ctx, FullHttpRequest request) {
            String requestId = request.headers().get("ww-req-id");
            if(requestId==null || requestId.isEmpty()) {
                requestId = UUID.randomUUID().toString().replace("-", "");
                request.headers().set("ww-req-id", requestId);
            }

            RequestContext context = new RequestContext(ctx, requestId, SvrConfig.process_in_iothread);
            RequestProcessor processor = new RequestProcessor();
            processor.process(context, request);
        }

        @Override
        public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws   Exception {
            SpecUtil.WWSpecLogError("server exception: " + cause.getStackTrace());
            cause.printStackTrace();
            ctx.close();
        }
    }

    /*
     * @usage 私有构造函数，防止被实例化
     */
    private NetworkService() {}
}
