package mytype.mycom.mygroup;

import io.netty.channel.ChannelFuture;
import io.netty.handler.codec.http.*;

import java.util.Arrays;
import java.util.concurrent.*;

import java.util.HashMap;
import java.util.Map;

import com.tencent.wework.SpecCallbackSDK;
import com.tencent.wework.SpecUtil;

import io.netty.buffer.Unpooled;
import io.netty.util.CharsetUtil;

/*
 * @usage 请求包处理器
 * 支持使用自定义线程池处理请求包
 *
 * */

public class RequestProcessor {
    /*
     * @description 处理POST请求的路径
     */
    private static final String POST_CONTEXT = "/";

    public RequestProcessor() {
    }

    /*
     * @usage 验证http请求是否合法
     */
    private HttpResponseStatus isValidRequest(FullHttpRequest request) {
        // 检查请求是否按照HTTP协议规范发送
        if (request.decoderResult().isFailure()) {
            return HttpResponseStatus.BAD_REQUEST;
        }
        // 检查请求方法是否为POST
        if (!HttpMethod.POST.equals(request.method())) {

            return HttpResponseStatus.METHOD_NOT_ALLOWED;
        }

        // 检查请求路径是否为"/"
        if (!request.uri().equals(POST_CONTEXT)) {
            return HttpResponseStatus.NOT_FOUND;
        }
        return HttpResponseStatus.OK;
    }

    /*
     * @param response http回包结构
     * @usage 通过io线程发送回包
     */
    public void sendResponse(RequestContext ctx, FullHttpResponse response) {
        long beginWriteTime = System.currentTimeMillis();
        ChannelFuture future = ctx.getChannelHandlerContext().writeAndFlush(response);
        future.addListener(writeFunc -> {
            // 打印请求处理耗时、等待耗时（从io线程收到请求到业务线程开始执行）、以及写回包耗时
            long writeTimeMs = System.currentTimeMillis() - beginWriteTime;
            if (!writeFunc.isSuccess()) {
                SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), "write err cost" + writeTimeMs);
            }
            else if(writeTimeMs > 10) {
                SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), "write cost  " + writeTimeMs + " ms");
            }
        });
    }

    /*
     * @return FullHttpResponse
     * @usage 构造出错时回包
     */
    private FullHttpResponse buildErrResponse(HttpResponseStatus errStatus) {
        return new DefaultFullHttpResponse(HttpVersion.HTTP_1_1, errStatus);
    }

    /*
     * @param ctx 请求上下文
     * @param request http请求体FullHttpRequest
     * @usage 将请求投递到业务线程异池步处理
     */
    public void process(RequestContext ctx, FullHttpRequest request) {

        //校验请求是否合法
        HttpResponseStatus status = isValidRequest(request);
        if(status != HttpResponseStatus.OK) {
            SpecUtil.WWSpecLogErrorWithReqId(ctx.getReqid(), "invalid http method: " + request.method() +  "status:" + status.code());
            FullHttpResponse response = buildErrResponse(status);
            sendResponse(ctx, response);
            ctx.getChannelHandlerContext().close();
            return;
        }

        if(ctx.isProcessInIoThread())
        {
            new ProcessTask(ctx, request).run();
        }
        else {
            try {
                request.retain();
                ctx.setAddTaskTime(System.currentTimeMillis());
                ThreadPoolSingleton.getInstance().execute(new ProcessTask(ctx, request));

            } catch (RejectedExecutionException e) {
                SpecUtil.WWSpecLogErrorWithReqId(ctx.getReqid(), "reject:queue is full reject " + Arrays.toString(e.getStackTrace()));
                FullHttpResponse response = buildErrResponse(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                sendResponse(ctx, response);
                ctx.getChannelHandlerContext().close();
            }
        }
    }

    class ProcessTask implements Runnable {

        RequestContext ctx;
        FullHttpRequest req;

        public ProcessTask(RequestContext ctx, FullHttpRequest req) {
            this.ctx = ctx;
            this.req = req;
        }

        @Override
        public void run() {
            long processStartTime = System.currentTimeMillis();
            try {
                RequestHandler handler= new RequestHandler();
                handler.process(ctx, req);
            } catch (Throwable e) {
                SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), " exception " + Arrays.toString(e.getStackTrace()));
            } finally {
                if(!ctx.isProcessInIoThread()) {
                    req.release();
                }

                long currentTimeMillis = System.currentTimeMillis();
                long total_cost = currentTimeMillis - ctx.getStartReadTime();
                long processTimeMs = currentTimeMillis - processStartTime;

                if(ctx.isProcessInIoThread()){
                    SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), "run total " + total_cost +
                            " ms handle " + processTimeMs +  " ms");
                }
                else {
                    long wait_cost = processStartTime - ctx.getAddTaskTime();
                    SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), "run total " + total_cost + " ms wait " +
                            wait_cost + " ms handle " + processTimeMs + " ms");
                }
            }
        }
    }

    /*
     * @usage 请求包处理逻辑类
     * 包括解包、调用业务逻辑、发送回包
     */
    class  RequestHandler{

        /*
         * @return SDK的SpecCallbackSDK结构
         * @usage 将netty的FullHttpRequest结构解析为SDK的SpecCallbackSDK结构
         */
        private SpecCallbackSDK parseRequest(RequestContext ctx, FullHttpRequest request) {
            Map<String, String> requestHeaders = new HashMap<>(request.headers().size());
            for (Map.Entry<String, String> entry : request.headers().entries()) {
                requestHeaders.put(entry.getKey(), entry.getValue());
            }

            //SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(),"req header: " + requestHeaders);

            // 使用UTF-8编码解析请求体，请求体为加密后经过base64编码的形式
            String requestBody = request.content().toString(CharsetUtil.UTF_8);

            //SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(),"req body: " + requestBody);
            return new SpecCallbackSDK("POST", requestHeaders, requestBody);
        }

        /*
         * @return FullHttpResponse
         * @usage 构造回包
         */
        private FullHttpResponse buildFullResponse(SpecCallbackSDK callback, String responseContent) {

            // 构造响应头，加密响应体
            callback.BuildResponseHeaderBody(responseContent);
            Map<String, String> responseHeaders = callback.GetResponseHeaders();
            String responseBody = callback.GetResponseBody();
            //SpecUtil.WWSpecLogInfo("response build done, response headers: " + responseHeaders);

            // 返回响应
            FullHttpResponse response = new DefaultFullHttpResponse(
                    HttpVersion.HTTP_1_1,
                    HttpResponseStatus.OK,
                    Unpooled.copiedBuffer(responseBody, CharsetUtil.UTF_8)
            );
            response.headers().set(HttpHeaderNames.CONTENT_LENGTH, response.content().readableBytes());
            for (Map.Entry<String, String> header : responseHeaders.entrySet()) {
                response.headers().set(header.getKey(), header.getValue());
            }

            return response;
        }

        /*
         * @usage 执行用户的业务逻辑
         * 先解包，然后根据请求的类型分别调用业务自己实现的业务逻辑类，然后返回响应
         * 请求类型说明：1-应用主动调用，2-专区后台通知
         */
        public void process(RequestContext ctx, FullHttpRequest request) {

            SpecCallbackSDK callback = parseRequest(ctx, request);
            String responseContent;

            // 解包校验签名等失败
            if (!callback.IsOk()) {
                responseContent = CommonUtils.getErrorResponse("callback is not ok");
            }
            else
            {
                //打印解密后的请求参数
                SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(),
                        "req args corpId = " + callback.GetCorpId(),
                        "agentId = " + callback.GetAgentId(),
                        "callType = " + callback.GetCallType(),
                        "isAsync = " + callback.GetIsAsync(),
                        "jobInfo = " + callback.GetJobInfo(),
                        "abilityId = " + callback.GetAbilityId(),
                        "notifyId = " + callback.GetNotifyId(),
                        "data = " + callback.GetData());


                //根据callType来分别调用不同的用户业务逻辑处理类
                switch ((int) callback.GetCallType()) {
                    // 应用主动调用专区程序，详见：https://developer.work.weixin.qq.com/document/53336
                    case 1: {
                        DemoCallProgramHandler callProgramHandler = new DemoCallProgramHandler();
                        responseContent = callProgramHandler.process(callback);
                        break;

                    }
                    // 专区后台通知专区程序，详见：https://developer.work.weixin.qq.com/document/53419
                    case 2: {
                        DemoReceiveCallBackHandler receiveCallBackHandler = new DemoReceiveCallBackHandler();
                        responseContent = receiveCallBackHandler.process(callback);
                        break;
                    }
                    default:
                        responseContent = CommonUtils.getErrorResponse("unknown call type: " + callback.GetCallType());
                        break;
                }
            }

            SpecUtil.WWSpecLogInfoWithReqId(ctx.getReqid(), "sdk_ver:"+ SpecUtil.GetSDKVersion(), "sz "+ responseContent.length() + " rsp " + responseContent);
            FullHttpResponse response = buildFullResponse(callback, responseContent);
            RequestProcessor.this.sendResponse(ctx, response);
        }
    }
}

