package mytype.mycom.mygroup;

import com.tencent.wework.SpecUtil;

import java.lang.management.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * @usage 收集服务器指标，输出到日志
 * @note  1. 监控日志同样需要使用SpecUtil输出，才能被日志管理平台收集
 *        2. 性能监控日志应结合
 */
public class ResourceMonitor {

    /**
     * @description 监控指标收集器
     */
    private final MemoryMXBean memoryMXBean;
    private final List<GarbageCollectorMXBean> gcBeans;
    private final ThreadMXBean threadMXBean;
    private final ClassLoadingMXBean classLoadingMXBean;
    private final OperatingSystemMXBean osMXBean;

    /**
     * @description 收集监控指标用到的常量
     */
    private static final int INITIAL_DELAY = 0;
    private   int OUTPUT_PERIOD = 10;
    private static final double MB_SIZE = 1024.0 * 1024.0;
    private static final int WAIT_THREADS_TIME = 100;

    /**
     * @description 日志格式
     */
    private static final String MEMORY_LOG_FORMAT = "MEM(MB) Heap: Init: %.2f, Max: %.2f, Used: %.2f, Committed: %.2f " +
                                                    "Non-Heap: Init: %.2f, Max: %.2fMB, Used: %.2f, Committed: %.2f ";
    private static final String GC_LOG_FORMAT = "GC : [%s], Cnt: %d, Time: %dms ";
    private static final String THREAD_LOG_FORMAT = "ThreadCnt: %d, busy IO: %d Worker: %d. ";
    private static final String CLASS_LOADING_LOG_FORMAT = "ClassCnt Total: %d, Loaded: %d, Unloaded: %d ";
    private static final String OS_LOG_FORMAT = "Cpu Load: %.2f ";

    /**
     * @usage 实例化JMX监控对象
     */
    public ResourceMonitor(int monitor_interval) {
        this.memoryMXBean = ManagementFactory.getMemoryMXBean();
        this.gcBeans = ManagementFactory.getGarbageCollectorMXBeans();
        this.threadMXBean = ManagementFactory.getThreadMXBean();
        this.classLoadingMXBean = ManagementFactory.getClassLoadingMXBean();
        this.osMXBean = ManagementFactory.getOperatingSystemMXBean();
        this.OUTPUT_PERIOD = monitor_interval;

    }

    /**
     * @usage 启动输出监控日志的定时任务
     */
    public void start() {
        ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
        scheduler.scheduleAtFixedRate(() -> {
            try {
                SpecUtil.WWSpecLogInfo(collectData());
            } catch (Exception e) {
                SpecUtil.WWSpecLogError("collect data fail", e.getMessage());
                e.printStackTrace();
            }
        }, INITIAL_DELAY, OUTPUT_PERIOD, TimeUnit.SECONDS);        
    }

    /**
     * @usage 收集监控数据，考虑到日志会被刷新，这里将输出部分静态的配置信息
     * @return JMX能监控的数据
     */
    private String collectData() {
        StringBuffer sb = new StringBuffer();

        // 获取负载
        sb.append(String.format(OS_LOG_FORMAT,
                osMXBean.getSystemLoadAverage()
        ));

        // 获取线程信息
        sb.append(String.format(THREAD_LOG_FORMAT,
                threadMXBean.getThreadCount(),
                getActiveThreadCount("nioEventLoopGroup"),
                SvrConfig.process_in_iothread?0:ThreadPoolSingleton.getInstance().getActiveCount()
        ));

        // 获取内存使用情况
        MemoryUsage heapMemoryUsage = memoryMXBean.getHeapMemoryUsage();
        MemoryUsage nonHeapMemoryUsage = memoryMXBean.getNonHeapMemoryUsage();        
        sb.append(String.format(MEMORY_LOG_FORMAT, 
            bytesToMb(heapMemoryUsage.getInit()),
            bytesToMb(heapMemoryUsage.getMax()),
            bytesToMb(heapMemoryUsage.getUsed()),
            bytesToMb(heapMemoryUsage.getCommitted()),
            bytesToMb(nonHeapMemoryUsage.getInit()),
            bytesToMb(nonHeapMemoryUsage.getMax()),
            bytesToMb(nonHeapMemoryUsage.getUsed()),
            bytesToMb(nonHeapMemoryUsage.getCommitted())
        ));

        // 获取垃圾回收信息
        for (GarbageCollectorMXBean gcBean : gcBeans) {
            sb.append(String.format(GC_LOG_FORMAT,
                gcBean.getName(),
                gcBean.getCollectionCount(),
                gcBean.getCollectionTime()
            ));
        }


        // 获取类加载信息
        sb.append(String.format(CLASS_LOADING_LOG_FORMAT,
            classLoadingMXBean.getTotalLoadedClassCount(),
            classLoadingMXBean.getLoadedClassCount(),
            classLoadingMXBean.getUnloadedClassCount()
        ));



        return sb.toString();
    }

    private static double bytesToMb(long bytes) {
        return bytes / MB_SIZE;
    }

    /**
     * @usage 获取指定线程名前缀的活跃线程数
     */
    private int getActiveThreadCount(String threadNamePrefix) {
        if (!threadMXBean.isThreadCpuTimeSupported()) {
            return -1;
        }

        // 收集netty线程的累计运行时间，单位纳秒
        Map<Long, Long> nettyThreadRunningTime = new HashMap<>(threadMXBean.getThreadCount());
        for (long threadId : threadMXBean.getAllThreadIds()) {
            ThreadInfo threadInfo = threadMXBean.getThreadInfo(threadId);
            if (threadInfo != null && threadInfo.getThreadName().startsWith(threadNamePrefix)) {
                if (threadMXBean.getThreadCpuTime(threadId) > 0) {
                    nettyThreadRunningTime.put(threadId, threadMXBean.getThreadCpuTime(threadId));
                }
            }
        }

        // 让出CPU等待，确保线程的CPU时间已经更新
        try {
            TimeUnit.MILLISECONDS.sleep(WAIT_THREADS_TIME);
        } catch (InterruptedException e) {
            SpecUtil.WWSpecLogError("Failed to sleep", e.getMessage());
            e.printStackTrace();
            return -1;
        }

        // 通过检查CPU时间是否增长来判断线程是否活跃
        int activeThreadCount = 0;
        for (long threadId : nettyThreadRunningTime.keySet()) {
            if (threadMXBean.getThreadCpuTime(threadId) > nettyThreadRunningTime.get(threadId)) {
                activeThreadCount++;
            }
        }
        return activeThreadCount;
    }
}
